@@ -2,6 +2,7 @@ use core::fmt;
 
 use musli::error::Error;
 
+use crate::reader::integer;
 use crate::reader::Token;
 
 /// The span of an error.
@@ -51,20 +52,24 @@ pub(crate) enum ParseErrorKind {
     ExpectedOpenBrace(Token),
     ExpectedCloseBrace(Token),
     ExpectedOpenBracket(Token),
+    ExpectedCloseBracket(Token),
     InvalidEscape,
     BufferUnderflow,
     BufferOverflow,
     UnexpectedHexEscapeEnd,
     InvalidUnicode,
-    NumericalOverflow,
-    ExpectedWholeNumber,
     InvalidNumeric,
+    CharEmptyString,
     ExpectedNull,
     ExpectedTrue,
     ExpectedFalse,
     ExpectedBool(Token),
+    ExpectedString(Token),
     ExpectedValue(Token),
     ParseFloat(lexical::Error),
+    IntegerError(integer::Error),
+    #[cfg(feature = "std")]
+    ValueError(musli_value::ValueError),
     Eof,
     Custom,
 }
@@ -130,6 +135,9 @@ impl fmt::Display for ParseError {
             ParseErrorKind::ExpectedOpenBracket(actual) => {
                 write!(f, "expected opening bracket, found {actual} (at {span})")
             }
+            ParseErrorKind::ExpectedCloseBracket(actual) => {
+                write!(f, "expected closing bracket, found {actual} (at {span})")
+            }
             ParseErrorKind::InvalidEscape => write!(f, "invalid string escape (at {span})"),
             ParseErrorKind::BufferUnderflow => write!(f, "buffer underflow (at {span})"),
             ParseErrorKind::BufferOverflow => write!(f, "buffer overflow (at {span})"),
@@ -137,21 +145,32 @@ impl fmt::Display for ParseError {
                 write!(f, "unexpected end of hex escape (at {span})")
             }
             ParseErrorKind::InvalidUnicode => write!(f, "invalid unicode (at {span})"),
-            ParseErrorKind::NumericalOverflow => write!(f, "numerical overflow (at {span})"),
-            ParseErrorKind::ExpectedWholeNumber => write!(f, "expected whole number (at {span})"),
             ParseErrorKind::InvalidNumeric => write!(f, "not numeric (at {span})"),
+            ParseErrorKind::CharEmptyString => {
+                write!(f, "expected string with a single character (at {span})")
+            }
             ParseErrorKind::ExpectedNull => write!(f, "expected `null` (at {span})"),
             ParseErrorKind::ExpectedTrue => write!(f, "expected `true` (at {span})"),
             ParseErrorKind::ExpectedFalse => write!(f, "expected `false` (at {span})"),
             ParseErrorKind::ExpectedBool(actual) => {
                 write!(f, "expected boolean, found {actual} (at {span})")
             }
+            ParseErrorKind::ExpectedString(actual) => {
+                write!(f, "expected string, found {actual} (at {span})")
+            }
             ParseErrorKind::ExpectedValue(actual) => {
                 write!(f, "expected value, found {actual} (at {span})")
             }
             ParseErrorKind::ParseFloat(error) => {
                 write!(f, "expected float, got {error} (at {span})")
             }
+            ParseErrorKind::IntegerError(error) => {
+                write!(f, "expected integer, got {error} (at {span})")
+            }
+            #[cfg(feature = "std")]
+            ParseErrorKind::ValueError(error) => {
+                write!(f, "value error: {error} (at {span})")
+            }
             ParseErrorKind::Eof => write!(f, "eof while parsing (at {span})"),
             ParseErrorKind::Custom => {
                 #[cfg(feature = "std")]
@@ -198,3 +217,14 @@ impl Error for ParseError {
 
 #[cfg(feature = "std")]
 impl std::error::Error for ParseError {}
+
+#[cfg(feature = "std")]
+impl From<musli_value::ValueError> for ParseError {
+    fn from(error: musli_value::ValueError) -> Self {
+        ParseError {
+            span: Span::empty(),
+            kind: ParseErrorKind::ValueError(error),
+            custom: None,
+        }
+    }
+}
